import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from einops import rearrange, repeat
from contextlib import nullcontext, contextmanager

@contextmanager
def no_param_grad(model):
    requires_grad_states = {}
    for name, param in model.named_parameters():
        requires_grad_states[name] = param.requires_grad
        param.requires_grad = False
    try:
        yield
    finally:
        for name, param in model.named_parameters():
            param.requires_grad = requires_grad_states[name]

class PCLayer(nn.Module):
    def __init__(self, f_module, b_module, f_loss_fn=F.mse_loss, b_loss_fn=F.mse_loss, device=torch.device('cpu')):
        super().__init__()
        self.f_module = f_module
        self.b_module = b_module
        self.f_loss_fn = f_loss_fn
        self.b_loss_fn = b_loss_fn
        self.device = device
        self.to(device)
    
    @torch.no_grad()
    def get_io_shapes(self, x=None, y=None):
        if x is not None: return self.f_module(x)
        elif y is not None: return self.b_module(y)
        else: raise ValueError("Either x or y must be provided")
    
    def forward_energy(self, x, y):
        """
        x -> p_y
        p_y, y -> loss
        """
        p_y = self.f_module(x)
        return self.f_loss_fn(p_y, y)

    def backward_energy(self, x, y):
        """
        p_x <- y
        p_x, x -> loss
        """
        p_x = self.b_module(y)
        return self.b_loss_fn(p_x, x)

class PCModel(nn.Module):
    def __init__(self, layers, max_its=1, min_energy=1e-1, energy_lr=1e-3, energy_optimizer_class=optim.SGD, energy_grad_clip_norm=None, device=torch.device('cpu')):
        super().__init__()
        self.layers = layers
        self.cached_x_shape = None
        self.cached_y_shape = None
        self.max_its = max_its
        self.min_energy = min_energy
        self.energy_lr = energy_lr
        self.energy_optimizer_class = energy_optimizer_class
        self.energy_grad_clip_norm = energy_grad_clip_norm
        self.device = device
        self.to(device)
    
    @torch.no_grad()
    def get_io_shapes(self, x=None, y=None):
        if x is not None:
            for layer in self.layers:
                x = layer.get_io_shapes(x, None)
            return x.shape
        elif y is not None:
            for layer in reversed(self.layers):
                y = layer.get_io_shapes(None, y)
            return y.shape
    
    @torch.no_grad()
    def update_io_shapes(self, x=None, y=None):
        if (x is not None and self.cached_x_shape == x.shape) or (y is not None and self.cached_y_shape == y.shape): return
        if x is not None:
            self.cached_x_shape = x.shape
            zero_x = torch.zeros_like(x, device=self.device)
            self.cached_y_shape = self.get_io_shapes(zero_x)
        elif y is not None:
            self.cached_y_shape = y.shape
            zero_y = torch.zeros_like(y, device=self.device)
            self.cached_x_shape = self.get_io_shapes(zero_y)
    
    @torch.no_grad()
    def forwad_state_init(self, x):
        state_tensors = []
        for layer in self.layers:
            state_tensors.append(x.clone().detach().requires_grad_(True))
            x = layer.f_module(x)
        state_tensors.append(x.clone().detach().requires_grad_(True))
        return state_tensors
    
    @torch.no_grad()
    def backward_state_init(self, y):
        state_tensors = []
        for layer in reversed(self.layers):
            state_tensors.append(y.clone().detach().requires_grad_(True))
            y = layer.b_module(y)
        state_tensors.append(y.clone().detach().requires_grad_(True))
        return state_tensors[::-1]

    def compute_model_energy(self, state_tensors):
        energy = torch.stack([
            layer.forward_energy(state_tensors[j], state_tensors[j+1])
            for j, layer in enumerate(self.layers)
        ]+[
            layer.backward_energy(state_tensors[j], state_tensors[j+1])
            for j, layer in enumerate(self.layers)
        ]).sum()
        return energy
    
    def train_forward(self, x, y, param_optimizer, param_grad_clip_norm=None, iterative=False, init_dir="forward"):
        self.update_io_shapes(x, y)
        
        # Amortized state initialization
        if init_dir == "forward":
            state_tensors = self.forwad_state_init(x)
            state_tensors[-1] = y
        elif init_dir == "backward":
            state_tensors = self.backward_state_init(y)
            state_tensors[0] = x
        else:
            raise ValueError("State initialization direction must be either 'forward' or 'backward'")
        
        energy_optimizer = self.energy_optimizer_class(state_tensors[1:-1], lr=self.energy_lr)
        
        if iterative:
            # Iterative training:
            # One energy convergence step per parameter update
            for i in range(self.max_its):
                energy_optimizer.zero_grad()
                param_optimizer.zero_grad()
                energy = self.compute_model_energy(state_tensors)
                energy.backward()
                if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                if param_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(self.parameters(), max_norm=param_grad_clip_norm)
                energy_optimizer.step()
                param_optimizer.step()
                if energy.item() < self.min_energy: break
        else:
            # Non-iterative training:
            # Energy convergence before parameter update
            param_optimizer.zero_grad()
            for i in range(self.max_its):
                energy_optimizer.zero_grad()
                energy = self.compute_model_energy(state_tensors)
                energy.backward()
                if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                if param_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(self.parameters(), max_norm=param_grad_clip_norm)
                energy_optimizer.step()
                if energy.item() < self.min_energy: break
            param_optimizer.step()
        return energy
    
    def forward(self, x=None, y=None):
        assert not (x is None and y is None), "Either x or y must be provided"
        
        self.update_io_shapes(x, y)

        with nullcontext() if self.training else no_param_grad(self):
            if x is not None:
                # Forward inference
                state_tensors = self.forwad_state_init(x)
                energy_optimizer = self.energy_optimizer_class(state_tensors[1:], lr=self.energy_lr) # Exclude input
                for i in range(self.max_its):
                    energy_optimizer.zero_grad()
                    energy = self.compute_model_energy(state_tensors)
                    energy.backward()
                    if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                    energy_optimizer.step()
                    if energy.item() < self.min_energy: break
                return state_tensors[-1]
            elif y is not None:
                # Backward inference
                state_tensors = self.backward_state_init(y)
                energy_optimizer = self.energy_optimizer_class(state_tensors[:-1], lr=self.energy_lr) # Exclude output
                for i in range(self.max_its):
                    energy_optimizer.zero_grad()
                    energy = self.compute_model_energy(state_tensors)
                    energy.backward()
                    if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                    energy_optimizer.step()
                    if energy.item() < self.min_energy: break
                return state_tensors[0]

# from torchvision import datasets, transforms
# from torch.utils.data import DataLoader

# device = torch.device('cpu')

# transform = transforms.Compose([transforms.ToTensor(),])
# train_set = datasets.MNIST(root='./data', train=True, download=True, transform=transform)
# test_set = datasets.MNIST(root='./data', train=False, download=True, transform=transform)
# train_loader = DataLoader(train_set, batch_size=64, shuffle=True)
# test_loader = DataLoader(test_set, batch_size=64, shuffle=False)

# model = PCModel(
#     layers=nn.ModuleList([
#         PCLayer(nn.Sequential(nn.Linear(784, 128, bias=False), nn.ReLU()), nn.Sequential(nn.Linear(128, 784, bias=False), nn.ReLU()), device=device),
#         PCLayer(nn.Linear(128, 10, bias=False), nn.Linear(10, 128, bias=False), f_loss_fn=F.cross_entropy, b_loss_fn=F.mse_loss, device=device)
#     ]),
#     max_its=2,
#     min_energy=1e-3,
#     energy_lr=1e-0,
#     energy_optimizer_class=optim.SGD,
#     device=device
# )

# param_optimizer = optim.AdamW(model.parameters(), lr=1e-3, weight_decay=0)

# def train():
#     model.train()
#     for batch_idx, (data, target) in enumerate(train_loader):
#         data = data.flatten(1).to(device)
#         target = target.to(device)
#         target = F.one_hot(target, num_classes=10).float()
#         train_energy = model.train_forward(data, target, param_optimizer, iterative=False, init_dir="forward")
#         if batch_idx % 100 == 0 and batch_idx > 0:
#             with no_param_grad(model):
#                 y = model(data)
#                 acc = (y.argmax(dim=1) == target.argmax(dim=1)).float().mean()
#                 loss = F.cross_entropy(y, target)
#                 print(f"Train epoch: {epoch}, batch: {batch_idx}", train_energy.item(), loss.item(), 100*acc.item())

# def test():
#     model.eval()
#     total_loss = 0
#     correct = 0
#     for data, target in test_loader:
#         data = data.flatten(1).to(device)
#         target = target.to(device)
#         y = model(data)
#         loss = F.cross_entropy(y, target)
#         total_loss += loss.item()
#         correct += (y.argmax(dim=1) == target).sum().item()

#     print("Test Loss:", total_loss / len(test_loader))
#     print("Test Accuracy:", 100 * correct / len(test_loader.dataset))

# epochs = 10
# for epoch in range(epochs):
#     train()
#     test()


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.linalg as LA
import torch.autograd as AG
import math
from einops import rearrange
import opt_einsum
from typing import Optional
from rotary_embedding_torch import RotaryEmbedding

class MultiheadAttention(nn.Module):
    """
    Vanilla Multihead Attention.
    Slight difference: the typical 1/sqrt(d_model) attention score scale is now a per head learnable parameter beta initialized at 1/sqrt(d_model).
    """
    def __init__(self, d_model, n_heads, dropout=0.0, bias=True, add_bias_kv=False, 
                 add_zero_attn=False, batch_first=False):
        super().__init__()
        self.d_model = d_model
        self.n_heads = n_heads
        self.dropout = dropout
        self.batch_first = batch_first
        self.d_head = d_model // n_heads
        
        assert self.d_head * n_heads == self.d_model, "d_model must be divisible by n_heads"
        
        # Linear projections for query, key, and value
        self.beta = nn.Parameter(torch.zeros(self.n_heads))
        self.beta._no_weight_decay = True
        self.q_proj = nn.Linear(d_model, d_model, bias=bias)
        self.k_proj = nn.Linear(d_model, d_model, bias=bias)
        self.v_proj = nn.Linear(d_model, d_model, bias=bias)
        self.out_proj = nn.Linear(d_model, d_model, bias=bias)
        
        # Optional bias for key and value
        if add_bias_kv:
            self.bias_k = nn.Parameter(torch.empty(1, 1, d_model))
            self.bias_v = nn.Parameter(torch.empty(1, 1, d_model))
        else:
            self.bias_k = self.bias_v = None
            
        self.add_zero_attn = add_zero_attn
        
        self._reset_parameters()
        
    def _reset_parameters(self):
        # Initialize projections using Xavier uniform
        nn.init.constant_(self.beta, 0.)
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.out_proj.weight)
        
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.)
            
        if self.bias_k is not None:
            nn.init.xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            nn.init.xavier_normal_(self.bias_v)
            
    def forward(self, x, key_padding_mask=None, need_weights=True, 
                attn_mask=None, average_attn_weights=True, rope=None):
        """
        Forward pass for the MultiheadAttention module.
        
        Args:
            query: Query embeddings of shape (seq_len_q, batch_size, d_model) or 
                  (batch_size, seq_len_q, d_model) if batch_first=True
            key: Key embeddings of shape (seq_len_k, batch_size, d_model) or
                 (batch_size, seq_len_k, d_model) if batch_first=True
            value: Value embeddings of shape (seq_len_v, batch_size, d_model) or
                   (batch_size, seq_len_v, d_model) if batch_first=True
            key_padding_mask: If provided, specified padding elements in the key will
                              be ignored by the attention. Shape: (batch_size, seq_len_k)
            need_weights: If True, returns attention weights in addition to attention output
            attn_mask: 2D or 3D mask that prevents attention to certain positions
            average_attn_weights: If True, returns averaged attention weights over heads
            
        Returns:
            attn_output: Attention output of shape (seq_len_q, batch_size, d_model) or
                         (batch_size, seq_len_q, d_model) if batch_first=True
            attn_output_weights: Attention weights of shape (batch_size, seq_len_q, seq_len_k)
                                 if need_weights=True, otherwise None
        """
        # Handle batch_first option
        if self.batch_first:
            x = x.transpose(0, 1)
        
        tgt_len, bsz, d_model = x.shape
        src_len = x.shape[0]
        
        # Apply linear projections
        q = self.q_proj(x)  # (tgt_len, batch_size, d_model)
        k = self.k_proj(x)  # (src_len, batch_size, d_model)
        v = self.v_proj(x)  # (src_len, batch_size, d_model)
        
        # Handle bias for key and value if present
        if self.bias_k is not None and self.bias_v is not None:
            k = torch.cat([k, self.bias_k.repeat(1, bsz, 1)])
            v = torch.cat([v, self.bias_v.repeat(1, bsz, 1)])
            if attn_mask is not None:
                attn_mask = F.pad(attn_mask, (0, 1))
            if key_padding_mask is not None:
                key_padding_mask = F.pad(key_padding_mask, (0, 1))
            src_len += 1
        
        # Add zero attention if requested
        if self.add_zero_attn:
            src_len += 1
            k = torch.cat([k, torch.zeros((1, bsz, d_model), dtype=k.dtype, device=k.device)], d_model=0)
            v = torch.cat([v, torch.zeros((1, bsz, d_model), dtype=v.dtype, device=v.device)], d_model=0)
            if attn_mask is not None:
                attn_mask = F.pad(attn_mask, (0, 1))
            if key_padding_mask is not None:
                key_padding_mask = F.pad(key_padding_mask, (0, 1))
        
        # Reshape q, k, v for multi-head attention
        q = q.contiguous().view(tgt_len, bsz * self.n_heads, self.d_head).transpose(0, 1)
        k = k.contiguous().view(src_len, bsz * self.n_heads, self.d_head).transpose(0, 1)
        v = v.contiguous().view(src_len, bsz * self.n_heads, self.d_head).transpose(0, 1)
        
        if rope:
            q = rope.rotate_queries_or_keys(q.reshape(bsz, self.n_heads, tgt_len, self.d_head)).reshape(bsz * self.n_heads, tgt_len, self.d_head)
            k = rope.rotate_queries_or_keys(k.reshape(bsz, self.n_heads, src_len, self.d_head)).reshape(bsz * self.n_heads, src_len, self.d_head)
        
        # Calculate attention scores
        q = q / (math.sqrt(self.d_head) * torch.exp(self.beta).reshape(self.n_heads, 1, 1).repeat(bsz, 1, 1))
        attn_output_weights = torch.bmm(q, k.transpose(1, 2))  # (bsz * n_heads, tgt_len, src_len)
        
        # Apply attention mask if provided
        if attn_mask is not None:
            if attn_mask.dim() == 2:
                attn_mask = attn_mask.unsqueeze(0)
            elif attn_mask.dim() == 3:
                attn_mask = attn_mask.repeat(self.n_heads, 1, 1)
            attn_output_weights = attn_output_weights + attn_mask
        
        # Apply key padding mask if provided
        if key_padding_mask is not None:
            attn_output_weights = attn_output_weights.view(bsz, self.n_heads, tgt_len, src_len)
            attn_output_weights = attn_output_weights.masked_fill(
                key_padding_mask.unsqueeze(1).unsqueeze(2),
                float('-inf'),
            )
            attn_output_weights = attn_output_weights.view(bsz * self.n_heads, tgt_len, src_len)
        
        # Convert attention weights to probabilities
        attn_output_weights = F.softmax(attn_output_weights, dim=-1)
        attn_output_weights = F.dropout(attn_output_weights, p=self.dropout, training=self.training)
        
        # Apply attention weights to values
        attn_output = torch.bmm(attn_output_weights, v)  # (bsz * n_heads, tgt_len, d_head)
        
        # Reshape output
        attn_output = attn_output.transpose(0, 1).contiguous().view(tgt_len, bsz, d_model)
        attn_output = self.out_proj(attn_output)
        
        # Process attention weights if needed
        if need_weights:
            attn_output_weights = attn_output_weights.view(bsz, self.n_heads, tgt_len, src_len)
            if average_attn_weights:
                attn_output_weights = attn_output_weights.mean(dim=1)
        else:
            attn_output_weights = None
        
        # Return in the correct format depending on batch_first
        if self.batch_first:
            return attn_output.transpose(0, 1), attn_output_weights
        return attn_output, attn_output_weights

class LinearAttention(nn.Module):
    """
    Vanilla Linear Attention.
    Kernel function is elu + 1.
    """
    def __init__(self, d_model, n_heads, bias=True):
        super().__init__()
        self.d_model = d_model
        self.n_heads = n_heads
        self.d_head = d_model // n_heads
        self.q_proj = nn.Linear(d_model, d_model, bias=bias)
        self.k_proj = nn.Linear(d_model, d_model, bias=bias)
        self.v_proj = nn.Linear(d_model, d_model, bias=bias)
        self.out_proj = nn.Linear(d_model, d_model, bias=bias)
        
        self._reset_parameters()
    
    def _reset_parameters(self):
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.out_proj.weight)
        
        if self.q_proj.bias is not None:
            nn.init.constant_(self.q_proj.bias, 0.)
        if self.k_proj.bias is not None:
            nn.init.constant_(self.k_proj.bias, 0.)
        if self.v_proj.bias is not None:
            nn.init.constant_(self.v_proj.bias, 0.)
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.)
    
    def forward(self, x, rope=None, causal=True):
        bsz, seq_len, d_model = x.size()
        q = rearrange(self.q_proj(x), 'b s (h d) -> b h s d', h=self.n_heads)
        k = rearrange(self.k_proj(x), 'b s (h d) -> b h s d', h=self.n_heads)
        v = rearrange(self.v_proj(x), 'b s (h d) -> (b h) s d', h=self.n_heads).contiguous()
        
        if rope:
            q = rope.rotate_queries_or_keys(q).reshape(bsz*self.n_heads, seq_len, self.d_head).contiguous()
            k = rope.rotate_queries_or_keys(k).reshape(bsz*self.n_heads, seq_len, self.d_head).contiguous()
        
        # q = torch.exp(q)
        # k = torch.exp(k)
        q = F.elu(q) + 1
        k = F.elu(k) + 1
        
        if causal:
            kv = torch.cumsum(torch.matmul(k.unsqueeze(-1), v.unsqueeze(-2)), dim=1)
            k1 = torch.cumsum(k, dim=1)
        else:
            kv = torch.einsum('zsD, zsd -> zDd', k, v).unsqueeze(1)
            k1 = k.sum(dim=1, keepdim=True)
        out = torch.matmul(q.unsqueeze(-2), kv).squeeze(-2) / (q*k1).sum(-1, keepdim=True)
        
        out = rearrange(out, '(b h) s d -> b s (h d)', h=self.n_heads)
        return self.out_proj(out)

class OrthoLinearAttention(nn.Module):
    """
    Orthogonal Linear Attention.
    A derivative of linear attention that orthogonalizes queries and keys for each head to reduce crossterm interference.
    """
    def __init__(self, d_model: int, n_heads: int, bias: bool = True):
        super().__init__()
        self.d_model = d_model
        self.n_heads = n_heads
        self.d_head = d_model // n_heads
        
        self.beta = nn.Parameter(torch.zeros(self.n_heads))
        self.beta._no_weight_decay = True
        self.q_proj = nn.Linear(d_model, d_model, bias=bias)
        self.k_proj = nn.Linear(d_model, d_model, bias=bias)
        self.v_proj = nn.Linear(d_model, d_model, bias=bias)
        self.out_proj = nn.Linear(d_model, d_model, bias=bias)
        
        self._reset_parameters()
        
    def _reset_parameters(self):
        nn.init.constant_(self.beta, 0.)
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.out_proj.weight)
        
        if self.q_proj.bias is not None:
            nn.init.constant_(self.q_proj.bias, 0.)
        if self.k_proj.bias is not None:
            nn.init.constant_(self.k_proj.bias, 0.)
        if self.v_proj.bias is not None:
            nn.init.constant_(self.v_proj.bias, 0.)
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.)
    
    def forward(self, x: torch.Tensor, rope: Optional[RotaryEmbedding] = None, causal: bool = True) -> torch.Tensor:
        """
        Args:
            x (torch.Tensor): Input sequence of shape (batch_size, seq_len, d_model).
            rope (Optional[RotaryEmbedding]): Optional RoPE encoder for rotating queries and keys.

        Returns:
            torch.Tensor: Output sequence of shape (batch_size, seq_len, d_model).
        """
        bsz, seq_len, d_model = x.size()
        q = rearrange(self.q_proj(x), 'b s (h d) -> b h s d', h=self.n_heads)
        k = rearrange(self.k_proj(x), 'b s (h d) -> b h s d', h=self.n_heads)
        v = rearrange(self.v_proj(x), 'b s (h d) -> (b h) s d', h=self.n_heads).contiguous()
        
        if rope:
            q = rope.rotate_queries_or_keys(q).reshape(bsz * self.n_heads, seq_len, self.d_head).contiguous()
            k = rope.rotate_queries_or_keys(k).reshape(bsz * self.n_heads, seq_len, self.d_head).contiguous()
        else:
            q = q.reshape(bsz * self.n_heads, seq_len, self.d_head).contiguous()
            k = k.reshape(bsz * self.n_heads, seq_len, self.d_head).contiguous()
        
        beta = torch.exp(self.beta).reshape(self.n_heads, 1, 1).repeat(bsz, 1, 1)
        q = (beta * q).softmax(-1)# * q.norm(dim=-1, keepdim=True)
        k = (beta * k).softmax(-1)# * k.norm(dim=-1, keepdim=True)
        
        if causal:
            kv = torch.cumsum(torch.matmul(k.unsqueeze(-1), v.unsqueeze(-2)), dim=1)
            kn = torch.cumsum(k, dim=1)
        else:
            kv = torch.einsum('zsD, zsd -> zDd', k, v).unsqueeze(1)
            kn = k.sum(1, keepdim=True)
        
        out = torch.matmul(q.unsqueeze(-2), kv).squeeze(-2) / (q * kn).sum(-1, keepdim=True)
        out = rearrange(out, '(b h) s d -> b s (h d)', h=self.n_heads)
        return self.out_proj(out)

class CompressionAttention(nn.Module):
    """
    Compression Attention.
    A derivative of softmax attention that compresses input sequences to a fixed length before expanding back to the original length.
    Achieved by two linear with sequence length attention operations.
    """
    def __init__(self, d_model, n_heads, mlp_dim, compressed_len, chunk_size=None,
                 dropout=0.0, bias=True, batch_first=False):
        super().__init__()
        self.d_model = d_model
        self.n_heads = n_heads
        self.mlp_dim = mlp_dim
        self.d_head = d_model // n_heads
        self.compressed_len = compressed_len
        self.chunk_size = chunk_size
        self.batch_first = batch_first
        self.dropout = dropout
        
        self.q_c = nn.Parameter(torch.empty(compressed_len, d_model))
        self.q_c._no_weight_decay = True
        self.beta = nn.Parameter(torch.zeros(self.n_heads))
        self.beta._no_weight_decay = True
        self.q_proj = nn.Linear(d_model, d_model, bias=bias)
        self.k_proj = nn.Linear(d_model, d_model, bias=bias)
        self.v_proj = nn.Linear(d_model, d_model, bias=bias)
        self.out_proj = nn.Linear(d_model, d_model, bias=bias)

        self._reset_parameters()
        
    def _reset_parameters(self):
        # Initialize projections using Xavier uniform
        nn.init.constant_(self.beta, 0.)
        nn.init.xavier_uniform_(self.q_c)
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.out_proj.weight)
        
        if self.q_proj.bias is not None:
            nn.init.constant_(self.q_proj.bias, 0.)
        if self.k_proj.bias is not None:
            nn.init.constant_(self.k_proj.bias, 0.)
        if self.v_proj.bias is not None:
            nn.init.constant_(self.v_proj.bias, 0.)
        if self.out_proj.bias is not None:
            nn.init.constant_(self.out_proj.bias, 0.)
    
    def forward(self, x, rope=None, causal=True, sequential=False):
        if sequential: return self.sequential_forward(x, rope, causal)
        
        # Handle batch_first option
        if self.batch_first:
            x = x.transpose(0, 1)
        
        cmprs_len = self.compressed_len
        _, bsz, d_model = x.shape
        src_len = x.shape[0]
        
        q_c = self.q_c.unsqueeze(1).repeat(1, bsz, 1)  # (compressed_len, d_model)
        q_s = self.q_proj(x)  # (src_len, batch_size, d_model)
        k_s = self.k_proj(x)  # (src_len, batch_size, d_model)
        v_s = self.v_proj(x)  # (src_len, batch_size, d_model)
        
        # Reshape for multi-head attention
        q_c = rearrange(q_c, 'c b (h d) -> (b h) c d', h=self.n_heads).contiguous()
        q_s = rearrange(q_s, 's b (h d) -> (b h) s d', h=self.n_heads).contiguous()
        k_s = rearrange(k_s, 's b (h d) -> (b h) s d', h=self.n_heads).contiguous()
        v_s = rearrange(v_s, 's b (h d) -> (b h) s d', h=self.n_heads).contiguous()
        
        if rope:
            if rope.use_xpos:
                q_s, k_s = rope.rotate_queries_and_keys(q_s.reshape(bsz, self.n_heads, src_len, self.d_head), k_s.reshape(bsz, self.n_heads, src_len, self.d_head))
            else:
                q_s = rope.rotate_queries_or_keys(q_s.reshape(bsz, self.n_heads, src_len, self.d_head))
                k_s = rope.rotate_queries_or_keys(k_s.reshape(bsz, self.n_heads, src_len, self.d_head))
            q_s = q_s.reshape(bsz * self.n_heads, src_len, self.d_head).contiguous()
            k_s = k_s.reshape(bsz * self.n_heads, src_len, self.d_head).contiguous()
        
        kv_s = torch.cat([k_s, v_s], dim=-1)  # (bsz * n_heads, src_len, 2*d_head)
        
        ### Compression self attention ###
        c_attn_weights = torch.bmm(q_c, k_s.transpose(1, 2))  # (bsz * n_heads, cmprs_len, src_len)
        
        if causal:
            # Manually perform softmax with cumulative sum for causal attention
            c_attn_weights = torch.exp(c_attn_weights - torch.max(c_attn_weights, dim=-1, keepdim=True).values)  # (bsz * n_heads, cmprs_len, src_len)
            c_attn_norm = torch.cumsum(c_attn_weights, dim=-1)  # (bsz * n_heads, cmprs_len, src_len)
            c_attn_weights = F.dropout(c_attn_weights, p=self.dropout, training=self.training)
        else:
            # Convert attention weights to probabilities
            c_attn_weights = F.softmax(c_attn_weights, dim=-1)
            c_attn_weights = F.dropout(c_attn_weights, p=self.dropout, training=self.training)
        
        ### Expansion self attention ###
        q_s = q_s / (math.sqrt(self.d_head) * torch.exp(self.beta).reshape(self.n_heads, 1, 1).repeat(bsz, 1, 1))
        
        if causal:
            # Calculate attention scores for compressed output
            kv_c = torch.cumsum((c_attn_weights.unsqueeze(-1) * kv_s.unsqueeze(1)), dim=2) / c_attn_norm.unsqueeze(-1)  # (bsz * n_heads, cmprs_len, src_len, 2*d_head)
            k_c, v_c = kv_c.split([self.d_head, self.d_head], dim=-1)  # (bsz * n_heads, cmprs_len, src_len, d_head)
            s_attn_weights = torch.einsum('zsd, zcsd -> zsc', q_s, k_c)  # (bsz * n_heads, src_len, cmprs_len)
            
            # Convert attention weights to probabilities
            s_attn_weights = F.softmax(s_attn_weights, dim=-1)
            s_attn_weights = F.dropout(s_attn_weights, p=self.dropout, training=self.training)
            
            # Apply attention weights to values
            s_attn_output = torch.einsum('zsc, zcsd -> zsd', s_attn_weights, v_c)  # (bsz * n_heads, src_len, d_head)
        else:
            # Calculate attention scores for compressed output
            kv_c = torch.bmm(c_attn_weights, kv_s)  # (bsz * n_heads, cmprs_len, 2*d_head)
            k_c, v_c = kv_c.split([self.d_head, self.d_head], dim=-1)  # (bsz * n_heads, cmprs_len, d_head)
            s_attn_weights = torch.bmm(q_s, k_c.transpose(1, 2))  # (bsz * n_heads, src_len, cmprs_len)
            
            # Convert attention weights to probabilities
            s_attn_weights = F.softmax(s_attn_weights, dim=-1)
            s_attn_weights = F.dropout(s_attn_weights, p=self.dropout, training=self.training)
            
            # Apply attention weights to values
            s_attn_output = torch.bmm(s_attn_weights, v_c)  # (bsz * n_heads, src_len, d_head)
        
        # Reshape output
        s_attn_output = s_attn_output.transpose(0, 1).contiguous().view(src_len, bsz, d_model)
        
        # Apply final projection
        s_attn_output = self.out_proj(s_attn_output)
        
        # Return in the correct format depending on batch_first
        if self.batch_first:
            return s_attn_output.transpose(0, 1)
        return s_attn_output

    def sequential_forward(self, x, rope=None, causal=True):        
        # Handle batch_first option
        if self.batch_first:
            x = x.transpose(0, 1)
        
        cmprs_len = self.compressed_len
        _, bsz, d_model = x.shape
        src_len = x.shape[0]
        
        assert src_len % self.chunk_size == 0, "src_len must be divisible by chunk_size"
        num_chunks = src_len // self.chunk_size
        
        q_c = self.q_c.unsqueeze(1).repeat(1, bsz, 1)  # (compressed_len, d_model)
        q_s = self.q_proj(x)  # (src_len, batch_size, d_model)
        k_s = self.k_proj(x)  # (src_len, batch_size, d_model)
        v_s = self.v_proj(x)  # (src_len, batch_size, d_model)
        
        # Reshape for multi-head attention
        q_c = rearrange(q_c, 'c b (h d) -> (b h) c d', h=self.n_heads).contiguous()
        q_s = rearrange(q_s, '(n t) b (h d) -> (b h) n t d', h=self.n_heads, t=self.chunk_size).contiguous()
        k_s = rearrange(k_s, '(n t) b (h d) -> (b h) n t d', h=self.n_heads, t=self.chunk_size).contiguous()
        v_s = rearrange(v_s, '(n t) b (h d) -> (b h) n t d', h=self.n_heads, t=self.chunk_size).contiguous()
        
        if rope:
            if rope.use_xpos:
                q_s, k_s = rope.rotate_queries_and_keys(rearrange(q_s, '(b h) n t d -> b h (n t) d', h=self.n_heads), rearrange(k_s, '(b h) n t d -> b h (n t) d', h=self.n_heads))
            else:
                q_s = rope.rotate_queries_or_keys(rearrange(q_s, '(b h) n t d -> b h (n t) d', h=self.n_heads))
                k_s = rope.rotate_queries_or_keys(rearrange(k_s, '(b h) n t d -> b h (n t) d', h=self.n_heads))
            q_s = rearrange(q_s, 'b h (n t) d -> (b h) n t d', n=num_chunks).contiguous()
            k_s = rearrange(k_s, 'b h (n t) d -> (b h) n t d', n=num_chunks).contiguous()
        
        ### Compression self attention ###
        if causal:
            s_k_state = torch.empty((bsz * self.n_heads, num_chunks, cmprs_len, self.chunk_size, self.d_head), device=x.device)
            s_v_state = torch.empty_like(s_k_state, device=x.device)
        else:
            s_k_state = torch.empty((bsz * self.n_heads, num_chunks, cmprs_len, self.d_head), device=x.device)
            s_v_state = torch.empty_like(s_k_state, device=x.device)
        
        c_k_state = torch.zeros((bsz*self.n_heads, cmprs_len, self.d_head), device=x.device)
        c_v_state = torch.zeros((bsz*self.n_heads, cmprs_len, self.d_head), device=x.device)
        for i in range(num_chunks):
            c_k_in = torch.cat([c_k_state, k_s[:, i]], dim=1)  # (bsz * n_heads, cmprs_len+chunk_size, d_head)
            c_v_in = torch.cat([c_v_state, v_s[:, i]], dim=1)  # (bsz * n_heads, cmprs_len+chunk_size, d_head)
            c_kv_in = torch.cat([c_k_in, c_v_in], dim=-1)
            c_attn_weights = torch.bmm(q_c, c_k_in.transpose(1, 2)) # (bsz * n_heads, cmprs_len, cmprs_len+chunk_size)
            if causal:
                # Manually perform softmax with cumulative sum for causal attention
                c_attn_weights = torch.exp(c_attn_weights - torch.max(c_attn_weights, dim=-1, keepdim=True).values)  # (bsz * n_heads, cmprs_len, cmprs_len+chunk_size)
                c_attn_norm = torch.cumsum(c_attn_weights, dim=-1)  # (bsz * n_heads, cmprs_len, cmprs_len+chunk_size)
                c_attn_weights = F.dropout(c_attn_weights, p=self.dropout, training=self.training)
                c_kv_state = (torch.cumsum((c_attn_weights.unsqueeze(-1) * c_kv_in.unsqueeze(1)), dim=2) / c_attn_norm.unsqueeze(-1))[:, :, cmprs_len:]  # (bsz * n_heads, cmprs_len, chunk_size, d_head)
                c_k_state, c_v_state = c_kv_state.split([self.d_head, self.d_head], dim=-1)  # (bsz * n_heads, cmprs_len, chunk_size, d_head)
                s_k_state[:, i] = c_k_state.clone()
                s_v_state[:, i] = c_v_state.clone()
                c_k_state = c_k_state[:, :, -1]  # (bsz * n_heads, cmprs_len, d_head)
                c_v_state = c_v_state[:, :, -1]  # (bsz * n_heads, cmprs_len, d_head)
            else:
                # Convert attention weights to probabilities
                c_attn_weights = F.softmax(c_attn_weights, dim=-1)
                c_attn_weights = F.dropout(c_attn_weights, p=self.dropout, training=self.training)
                c_k_state = torch.einsum('zcT, zTd -> zcd', c_attn_weights, c_k_in)  # (bsz * n_heads, cmprs_len, d_head)
                c_v_state = torch.einsum('zcT, zTd -> zcd', c_attn_weights, c_v_in)  # (bsz * n_heads, cmprs_len, d_head)
                s_k_state[:, i] = c_k_state.clone()
                s_v_state[:, i] = c_v_state.clone()
        
        ### Expansion self attention ###
        q_s = q_s / (math.sqrt(self.d_head) * torch.exp(self.beta).reshape(self.n_heads, 1, 1, 1).repeat(bsz, 1, 1, 1))
        q_s = q_s.reshape(-1, self.chunk_size, self.d_head)
        
        if causal:
            s_k_state = s_k_state.reshape(-1, cmprs_len, self.chunk_size, self.d_head)
            s_v_state = s_v_state.reshape(-1, cmprs_len, self.chunk_size, self.d_head)
            
            # Calculate attention scores for compressed output
            s_attn_weights = torch.einsum('Ztd, Zctd -> Ztc', q_s, s_k_state)  # (bsz * n_heads, src_len, cmprs_len)
            
            # Convert attention weights to probabilities
            s_attn_weights = F.softmax(s_attn_weights, dim=-1)
            s_attn_weights = F.dropout(s_attn_weights, p=self.dropout, training=self.training)
            
            # Apply attention weights to values
            s_attn_output = torch.einsum('Ztc, Zctd -> Ztd', s_attn_weights, s_v_state)  # (bsz * n_heads, src_len, d_head)
        else:
            q_s = q_s.reshape(-1, self.chunk_size, self.d_head)
            s_k_state = s_k_state.reshape(-1, cmprs_len, self.d_head)
            s_v_state = s_v_state.reshape(-1, cmprs_len, self.d_head)

            # Calculate attention scores for compressed output
            s_attn_weights = torch.einsum('Ztd, Zcd -> Ztc', q_s, s_k_state)  # (bsz * n_heads * num_chunks, chunk_size, cmprs_len)
            
            # Convert attention weights to probabilities
            s_attn_weights = F.softmax(s_attn_weights, dim=-1)
            s_attn_weights = F.dropout(s_attn_weights, p=self.dropout, training=self.training)
            
            # Apply attention weights to values
            s_attn_output = torch.einsum('Ztc, Zcd -> Ztd', s_attn_weights, s_v_state)  # (bsz * n_heads * num_chunks, chunk_size, d_head)
        
        # Reshape output
        s_attn_output = s_attn_output.reshape(bsz*self.n_heads, src_len, self.d_head).transpose(0, 1).contiguous().view(src_len, bsz, d_model)
        
        # Apply final projection
        s_attn_output = self.out_proj(s_attn_output)
        
        # Return in the correct format depending on batch_first
        if self.batch_first:
            return s_attn_output.transpose(0, 1)
        return s_attn_output



import os
import sys
import time
import yaml
import copy
import torch
import wandb
import traceback
from tqdm import tqdm
from torch.utils.data import DataLoader
from .utils import *
from .models import initialize_model
from .schedulers import initialize_scheduler
from .datasets import initialize_dataset

def train_epoch(epoch, train_loader, model, optimizer, loss_fn, acc_fn,
                scheduler=None, device=torch.device("cpu"),
                output_dir="", model_name=None,
                val_loader=None,
                wandb_logging=True, wandb_metrics=["acc", "loss"],
                grad_clip_norm=None, accumulation_steps=0,
                checkpoint_freq=None, val_freq=None, info_freq=None):
    # Default model name
    if model_name is None: model_name = model.__class__.__name__
    model.train()
    train_loss = 0
    correct = 0
    iterable = tqdm(train_loader, desc=f"Train Epoch {epoch}", leave=False, bar_format='{desc}: [{n_fmt}/{total_fmt}] {percentage:.0f}%|{bar}| [{rate_fmt}] {postfix}')
    for batch_idx, (data, target) in enumerate(iterable):
        # Forward pass
        data = data.to(device)
        target = target.to(device)
        optimizer.zero_grad()
        output = model(data)
        
        # Accuracy
        accuracy = acc_fn(output.clone().detach(), target.clone().detach())
        correct += accuracy
        
        # Loss
        loss = loss_fn(output, target)
        batch_loss = loss.item()
        train_loss += loss.item()
        
        # Backward pass and gradient accumulation if applicable
        loss = loss / (accumulation_steps + 1)
        loss.backward()
        
        if (batch_idx + 1) % (accumulation_steps + 1) == 0:
            if grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=grad_clip_norm)
            optimizer.step()
            optimizer.zero_grad()
            if scheduler: scheduler.step()
        
        # WandB logging
        if wandb_logging:
            log_data = {}
            if "acc" in wandb_metrics: log_data["train/acc"] = 100. * accuracy / len(data)
            if "loss" in wandb_metrics: log_data["train/loss"] = batch_loss
            if "ppl" in wandb_metrics: log_data["train/ppl"] = torch.exp(batch_loss)
            if "lr" in wandb_metrics: log_data["misc/lr"] = scheduler.get_last_lr()[0]
            if "seq_len" in wandb_metrics: log_data["misc/seq_len"] = train_loader.dataset.len
            wandb.log(log_data)
        
        # Post info
        if info_freq and batch_idx % info_freq == 0 and batch_idx != 0:
            tqdm.write(f'Train Epoch {epoch}: [{batch_idx}/{len(train_loader)}] LR: {scheduler.get_last_lr()[0]:.1e}, Loss: {batch_loss:.4f}, Acc: {100. * accuracy / len(data):.0f}%')
        
        # Checkpoint
        if checkpoint_freq and batch_idx % checkpoint_freq == 0 and batch_idx != 0:
            checkpoint(model_name, output_dir, model, optimizer, scheduler)
        
        # Validation
        if val_freq and batch_idx % val_freq == 0 and batch_idx != 0:
            if val_loader: val_epoch(model, val_loader, loss_fn, acc_fn, device=device, wandb_logging=wandb_logging)
            model.train()
    
    # Step sequence length if applicable
    try: train_loader.dataset.step()
    except: pass
    
    train_loss /= len(train_loader)
    train_acc = 100. * correct / len(train_loader.dataset)
    
    return train_loss, train_acc

@ torch.no_grad()
def val_epoch(model, val_loader, loss_fn, acc_fn,
              device=torch.device("cpu"),
              wandb_logging=True, wandb_metrics=["acc", "loss"],):
    model.eval()
    val_loss = 0
    correct = 0
    start = time.time()
    iterable = val_loader
    for data, target in iterable:
        data = data.to(device)
        target = target.to(device)
        output = model(data)
        val_loss += loss_fn(output, target).item()
        correct += acc_fn(output, target)

    total_time = time.time() - start
    val_loss /= len(val_loader)
    val_acc = 100. * correct / len(val_loader.dataset)
    
    tqdm.write(f'\033[93mVal Epoch: Loss: {val_loss:.4f}, Acc: {correct}/{len(val_loader.dataset)} ({val_acc:.0f}%), Elapsed: {total_time:.3f}s\033[0m')
    if wandb_logging:
        log_data = {}
        if "acc" in wandb_metrics: log_data["val/acc"] = val_acc
        if "loss" in wandb_metrics: log_data["val/loss"] = val_loss
        if "ppl" in wandb_metrics: log_data["val/ppl"] = torch.exp(val_loss)
        wandb.log(log_data)
    
    return val_loss, val_acc

@ torch.no_grad()
def test_epoch(model, test_loader, loss_fn, acc_fn,
               device=torch.device("cpu"),
               wandb_logging=True, wandb_metrics=["acc", "loss"],):
    model.eval()
    test_loss = 0
    correct = 0
    start = time.time()
    tqdm.write("")
    iterable = tqdm(test_loader, desc=f"Test Epoch", leave=False, bar_format='\033[92m{desc}: [{n_fmt}/{total_fmt}] {percentage:.0f}%|{bar}| [{rate_fmt}] {postfix}\033[0m')
    for data, target in iterable:
        data = data.to(device)
        target = target.to(device)
        output = model(data)
        test_loss += loss_fn(output, target).item()
        correct += acc_fn(output, target)

    total_time = time.time() - start
    test_loss /= len(test_loader)
    test_acc = 100. * correct / len(test_loader.dataset)
    
    tqdm.write(f'\033[92mTest Epoch: Loss: {test_loss:.4f}, Acc: {correct}/{len(test_loader.dataset)} ({test_acc:.0f}%), Elapsed: {total_time:.3f}s\033[0m\n')
    if wandb_logging:
        log_data = {}
        if "acc" in wandb_metrics: log_data["test/acc"] = test_acc
        if "loss" in wandb_metrics: log_data["test/loss"] = test_loss
        if "ppl" in wandb_metrics: log_data["test/ppl"] = torch.exp(test_loss)
        wandb.log(log_data)
        
    return test_loss, test_acc

def train(epochs, benchmark_name, model, train_loader, optimizer, loss_fn, acc_fn,
          scheduler=None, device=torch.device("cpu"),
          train_config=None,
          output_dir="", model_name=None,
          val_loader=None, test_loader=None,
          local_log_path=None,
          wandb_logging=True, wandb_entity=None, wandb_project=None,
          wandb_metrics=["acc", "loss"],
          grad_clip_norm=None, accumulation_steps=0,
          checkpoint_freq=None, val_freq=None, info_freq=None):
    try:
        sys.stdout.write("\033[?25l")
        
        NoEcho.disable_echo()
        
        # Default model name
        if model_name is None: model_name = model.__class__.__name__
        
        print(f'\033[1m{benchmark_name} Benchmark\033[0m')
        print(f'\033[1m{model_name}\033[0m')
        print(f'\033[4mTotal params: {count_parameters(model):,}\033[0m\n')
        
        # WandB Initialization
        if wandb_logging:
            assert wandb_entity is not None, "WandB entity is required for logging."
            assert wandb_project is not None, "WandB project is required for logging."
            wandb.init(
                settings=wandb.Settings(silent=True),
                mode="online" if online() else "offline",
                entity=wandb_entity,
                project=wandb_project,
                name=model_name,
                config=train_config,
            )
        
        # Use multiple GPUs if available
        if torch.cuda.device_count() > 1: model = nn.DataParallel(model)
        
        # Allocate dynamic memory if applicable for dataset
        if hasattr(train_loader.dataset, "seq_len_range"):
            min_len, max_len = train_loader.dataset.seq_len_range()
            model = allocate_dynamic_memory(model, train_loader.batch_size, min_len, max_len, device)
        
        # Metrics
        train_losses = []
        train_accuracies = []
        if test_loader:
            test_losses = []
            test_accuracies = []
        else:
            test_losses = None
            test_accuracies = None
        
        # Train loop
        for epoch in range(1, epochs + 1):
            # Train epoch
            train_loss, train_acc = train_epoch(
                epoch=epoch, train_loader=train_loader, val_loader=val_loader,
                model=model, optimizer=optimizer, scheduler=scheduler,
                loss_fn=loss_fn, acc_fn=acc_fn, device=device,
                output_dir=output_dir, model_name=model_name,
                wandb_logging=wandb_logging, wandb_metrics=wandb_metrics,
                grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq
            )
            train_losses.append(train_loss)
            train_accuracies.append(train_acc)
            
            # Test epoch
            if test_loader:
                test_loss, test_acc = test_epoch(
                    model, test_loader, loss_fn, acc_fn,
                    device=device,
                    wandb_logging=wandb_logging, wandb_metrics=wandb_metrics
                )
                test_losses.append(test_loss)
                test_accuracies.append(test_acc)
            
            # Model checkpoint
            checkpoint(model_name=model_name, output_dir=output_dir,
                    model=model, optimizer=optimizer, scheduler=scheduler)
        
        # Final logging
        if local_log_path: log_info(log_path=local_log_path, model=model, model_name=model_name, configs=train_config, train_accuracies=train_accuracies, test_accuracies=test_accuracies)
        wandb.finish()
    
    except KeyboardInterrupt as e: raise e
    except Exception as e:
        if wandb_logging: wandb.finish()
        raise e
    finally:
        NoEcho.enable_echo()
        if wandb_logging: cleanup_wandb(wandb_entity, wandb_project)
        sys.stdout.write("\033[?25h")

def train_from_config_file(yaml_path, loss_fn, acc_fn, device=torch.device("cpu")):
    """
    Config file options:
        global:
            benchmark_name: Name of the benchmark.
            output_dir: Directory to save model checkpoints.
            
            dataset:
                name: Class name of the dataset to use.
                splits: Dictionary of dataset splits with their configurations. (e.g., "train", "val", "test")
            
            logging (optional):
                info_freq (default: 100): Frequency of CLI logging training information. No CLI logging if unspecified.
                local_log_path: Path to save local logs.
                wandb: WandB logging configurations.
                    entity: WandB entity name.
                    project: WandB project name.
                    metrics (default: ["acc", "loss"]): List of metrics to log. (e.g., ["acc", "loss", "ppl", "lr", "seq_len"]).
            
            val_freq (default: 500): Frequency of validation during training. No validation if unspecified.
            checkpoint_freq (default: 500): Frequency of saving model checkpoints. No checkpointing if unspecified.
        
        experiments: List of experiments to run.\n
            Format:
                general:
                    seed (default: 0): Random seed for reproducibility.
                    batch_size (default: 32): Batch size for training.
                    accumulation_steps (default: 0): Number of batches to accumulate gradients for.
                    epochs (default: 1): Number of epochs to train.
                    grad_clip_norm (optional): Gradient clipping norm. No clipping if unspecified.
                    load_checkpoint (default: False): Whether to attempt loading model from checkpoint.
                
                model:
                    name: Class name of the model to use.
                    Other model-specific configurations.
                
                optimizer:
                    name: Optimizer class name (e.g., "Adam", "SGD").
                    Other optimizer-specific configurations.
                
                scheduler (optional):
                    name: Scheduler class name (e.g., "StepLR", "CosineAnnealingLR").
                    Other scheduler-specific configurations.

    Args:
        yaml_path (str): Path to YAML configuration file.
        loss_fn (Callable): A function to compute model loss. Takes model output and target as inputs.
        acc_fn (Callable): A function to compute model accuracy. Takes model output and target as inputs.
        device (torch.device, optional): Device to run the training on. Defaults to cpu.
    """
    os.system('clear')
    
    with open(yaml_path, 'r') as f:
        config = yaml.safe_load(f)
    
    # Extract global training configurations
    global_config = config.get("global")
    benchmark_name = global_config.get("benchmark_name")
    output_dir = global_config.get("output_dir")
    
    # Initialize datasets
    dataset_config = global_config.get("dataset")
    dataset_name = dataset_config.get("name")
    dataset_splits = dataset_config.get("splits", {})
    if "train" in dataset_splits:
        dataset_args = dataset_splits["train"]
        train_dataset = initialize_dataset(dataset_name, **dataset_args)
    val_dataset = None
    if "val" in dataset_splits:
        dataset_args = dataset_splits["val"]
        val_dataset = initialize_dataset(dataset_name, **dataset_args)
    test_dataset = None
    if "test" in dataset_splits:
        dataset_args = dataset_splits["test"]
        test_dataset = initialize_dataset(dataset_name, **dataset_args)
    
    # Get logging configurations
    logging_config = global_config.get("logging", {})
    info_freq = logging_config.get("info_freq", 100)
    local_log_path = logging_config.get("local_log_path")
    wandb_config = logging_config.get("wandb", {})
    wandb_logging = wandb_config is not None
    wandb_entity = wandb_config.get("entity")
    wandb_project = wandb_config.get("project")
    wandb_metrics = wandb_config.get("metrics", ["acc", "loss"])
    
    # Get checkpointing configurations
    val_freq = global_config.get("val_freq", 500)
    checkpoint_freq = global_config.get("checkpoint_freq", 500)
    
    # Run all experiments
    successful_count = 0
    experiments = config.get("experiments")
    for i, experiment in enumerate(experiments):
        print(f'\033[1mRunning Experiment [{i + 1}/{len(experiments)}]\033[0m\n')
        
        general_config = copy.deepcopy(experiment.get("general"))
        general_config = try_to_float(general_config)
        
        # Set seed
        seed = general_config.get("seed", 0)
        torch.manual_seed(seed)
        if device == torch.device("cuda"):
            torch.cuda.manual_seed_all(seed)
        
        # Initialize dataloaders
        batch_size = general_config.get("batch_size", 32)
        epochs = general_config.get("epochs", 1)
        grad_clip_norm = general_config.get("grad_clip_norm", None)
        accumulation_steps = general_config.get("accumulation_steps", 0)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
        val_loader = None
        test_loader = None
        if val_dataset: val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)
        if test_dataset: test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)
        
        model_config = copy.deepcopy(experiment.get("model"))
        model_name = model_config.pop("name")
        model_config = try_to_float(model_config)
        model_args = model_config.copy()
        model_args["device"] = device
        
        # Initialize model
        model = initialize_model(model_name, **model_args)
        
        # Initialize optimizer
        def initialize_optimizer(name, *args, **kwargs):
            optimizer_class = getattr(sys.modules["torch.optim"], name, None)
            return optimizer_class(*args, **kwargs)
        optimizer_config = copy.deepcopy(experiment.get("optimizer"))
        optimizer_name = optimizer_config.pop("name")
        optimizer_config = try_to_float(optimizer_config)
        weight_decay = float(optimizer_config.get("weight_decay", 0.0))
        optimizer_config["params"] = apply_weight_decay(model, weight_decay)
        optimizer = initialize_optimizer(optimizer_name, **optimizer_config)
        
        # Initialize scheduler if specified
        scheduler = None
        scheduler_config = copy.deepcopy(experiment.get("scheduler", {}))
        if scheduler_config:
            scheduler_name = scheduler_config.pop("name")
            scheduler_config = try_to_float(scheduler_config)
            scheduler_config["optimizer"] = optimizer
            scheduler = initialize_scheduler(scheduler_name, **scheduler_config)
        
        # Load model from checkpoint if specified
        load_from_checkpoint = general_config.get("load_checkpoint", False)
        if load_from_checkpoint:
            model, optimizer, scheduler = load_checkpoint(
                model_name=model_name, output_dir=output_dir,
                model=model, optimizer=optimizer, scheduler=scheduler,
                device=device
            )
        else: print(f'\033[91mStarting from scratch\033[0m')
        
        # Collect all training configurations for logging
        train_config = model_config.copy()
        train_config.update({
            "benchmark": benchmark_name,
            "model": model_name,
            "seed": seed,
            "bsz": batch_size,
            "accumulation_steps": accumulation_steps,
            "lr": optimizer_config.get("lr"),
            "warmup_epochs": scheduler_config.get("warmup_epochs", 0),
            "total_epochs": epochs,
            "weight_decay": weight_decay,
            "grad_clip_norm": grad_clip_norm,
            "permuted": dataset_splits["train"].get("permuted"),
        })
        
        # Train the model
        successful = True
        try:
            train(
                epochs=epochs, benchmark_name=benchmark_name, model_name=model_name,
                model=model, optimizer=optimizer, scheduler=scheduler,
                loss_fn=loss_fn, acc_fn=acc_fn,
                train_config=train_config,
                output_dir=output_dir,
                train_loader=train_loader, val_loader=val_loader, test_loader=test_loader,
                local_log_path=local_log_path,
                wandb_logging=wandb_logging, wandb_entity=wandb_entity, wandb_project=wandb_project,
                wandb_metrics=wandb_metrics,
                grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq,
                device=device,
            )
        except KeyboardInterrupt:
            successful = False
            terminate = input('Terminate all experiments? (y/n): ').strip().lower()
            if terminate == 'y': break
        except Exception as e:
            successful = False
            traceback_str = traceback.format_exc()
            print(f'\033[91mExperiment [{i + 1}/{len(experiments)}] failed with error:\n{traceback_str}\033[0m\n')
        if successful:
            successful_count += 1
            print(f'\033[92mExperiment [{i + 1}/{len(experiments)}] completed successfully\033[0m\n')
        
        del general_config, model_config, optimizer_config, scheduler_config, train_config
    
    print(f'\033[1m{successful_count}/{len(experiments)} experiments completed successfully\033[0m')


import torch
import torch.nn as nn
import torch.nn.functional as F

class RMSNorm(nn.Module):
    def __init__(self, d, eps=1e-5, device=None):
        """Gated Root Mean Square Layer Normalization

        Paper: https://arxiv.org/abs/1910.07467
        """
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(d, device=device))

    def forward(self, x, z=None):
        if z is not None:
            x = x * F.silu(z)
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps) * self.weight




import os
import sys
import torch
from torchvision import datasets, transforms
from torch.utils.data import Dataset
from PIL import Image
import random
import pandas as pd
from collections import defaultdict
from transformers import AutoTokenizer
from datasets import load_dataset

class SequentialMNIST(datasets.MNIST):
    def __init__(self, root, train, download=True, permuted=False):
        self.transform = [
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.view(-1, 1))
        ]
        if permuted:
            random_permutation = torch.randperm(28 * 28)
            self.transform.append(transforms.Lambda(lambda x: x.view(-1)[random_permutation].view(-1, 1)))
        self.transform = transforms.Compose(self.transform)
        super().__init__(root, train=train, download=download, transform=self.transform)

class SequentialEMNIST(datasets.EMNIST):
    def __init__(self, root, train, split, download=True, permuted=False):
        self.transform = [
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.view(-1, 1))
        ]
        if permuted:
            random_permutation = torch.randperm(28 * 28)
            self.transform.append(transforms.Lambda(lambda x: x.view(-1)[random_permutation].view(-1, 1)))
        self.transform = transforms.Compose(self.transform)
        super().__init__(root, train=train, download=download, split=split, transform=self.transform)

class SequentialFashionMNIST(datasets.FashionMNIST):
    def __init__(self, root, train, download=True, permuted=False):
        self.transform = [
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.view(-1, 1))
        ]
        if permuted:
            random_permutation = torch.randperm(28 * 28)
            self.transform.append(transforms.Lambda(lambda x: x.view(-1)[random_permutation].view(-1, 1)))
        self.transform = transforms.Compose(self.transform)
        super().__init__(root, train=train, download=download, transform=self.transform)

class SequentialCIFAR10(datasets.CIFAR10):
    def __init__(self, root, train, download=True, permuted=False):
        self.transform = [
            transforms.Grayscale(num_output_channels=1),
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.view(-1, 1))
        ]
        if permuted:
            random_permutation = torch.randperm(32 * 32)
            self.transform.append(transforms.Lambda(lambda x: x.view(-1)[random_permutation].view(-1, 1)))
        self.transform = transforms.Compose(self.transform)
        super().__init__(root, train=train, download=download, transform=self.transform)

class SequentialCIFAR100(datasets.CIFAR100):
    def __init__(self, root, train, download=True, permuted=False):
        self.transform = [
            transforms.Grayscale(num_output_channels=1),
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.view(-1, 1))
        ]
        if permuted:
            random_permutation = torch.randperm(32 * 32)
            self.transform.append(transforms.Lambda(lambda x: x.view(-1)[random_permutation].view(-1, 1)))
        self.transform = transforms.Compose(self.transform)
        super().__init__(root, train=train, download=download, transform=self.transform)

class Pathfinder(Dataset):
    def __init__(self, root, dim, train, subset="curv_baseline", split_idx=180, permuted=False):
        self.transform = [
            transforms.ToTensor(),
            transforms.Lambda(lambda x: x.view(-1, 1))
        ]
        if permuted:
            random_permutation = torch.randperm(dim**2)
            self.transform.append(transforms.Lambda(lambda x: x.view(-1)[random_permutation].view(-1, 1)))
        self.transform = transforms.Compose(self.transform)
        self.data = []
        
        min_idx = 0 if train else split_idx
        max_idx = split_idx if train else 200
        dataset_root = os.path.join(root, f"pathfinder{dim}", subset)
        metadata_root = os.path.join(root, f"pathfinder{dim}", subset, 'metadata')
        for i in range(min_idx, max_idx):
            with open(os.path.join(metadata_root, f"{i}.npy"), 'r') as file:
                for line in file:
                    parts = line.strip().split()
                    img_rel_path = parts[0:2]
                    label = int(parts[3])
                    img_path = os.path.join(dataset_root, *img_rel_path)
                    self.data.append((img_path, label))

    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        image = None
        while image is None:
            img_path, label = self.data[idx]
            try:
                image = Image.open(img_path).convert('RGB')
            except:
                idx = (idx + 1) % len(self.data)
        if self.transform:
            image = self.transform(image)
        return image, label

class ListOps(Dataset):
    def __init__(self, root, split, min_len=1, max_len=1000, warmup_epochs=0, balance=False):
        """
        splits: ["train", "val", "test"]
        """
        if warmup_epochs < 1:
            self.min_len = max_len
        else:
            self.min_len = min_len
        self.max_len = max_len
        self.len = self.min_len
        self.step_size = (self.max_len - self.min_len) // (warmup_epochs + 1)
        self.data = pd.read_csv(f"{root}/listops/basic_{split}.tsv", sep="\t")
        
        if balance: self._balance_data()
    
    def _balance_data(self):
        grouped = defaultdict(list)
        for idx, item in self.data.iterrows():
            grouped[item["Target"]].append(item)

        min_class_size = min(len(items) for items in grouped.values())

        balanced_items = []
        for items in grouped.values():
            balanced_items.extend(items[:min_class_size])

        self.data = pd.DataFrame(balanced_items)
        
    def tokenizer(self, data):
        token_map = {
            "CLS": 0,
            "PAD": 1,
            "[MAX": 2,
            "[MIN": 3,
            "[MED": 4,
            "[SM": 5,
            "]": 6,
            **{str(i): i + 7 for i in range(10)}
        }

        src = data["Source"].translate({ ord("("): None, ord(")"): None })
        tokens = ["CLS"] + src.split()
        try:
            tokenized = [token_map[token] for token in tokens]
        except KeyError as e:
            raise ValueError(f"Unexpected token: {e.args[0]}")
        
        tokenized = torch.tensor(tokenized, dtype=torch.long)
        target = torch.tensor(data["Target"], dtype=torch.long)
        return tokenized, target
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        item = self.data.iloc[idx]
        tokenized, target = self.tokenizer(item)
        padded_tokenized = torch.nn.functional.pad(tokenized, (0, self.len - tokenized.size(0)), value=1)  # Pad with PAD token (1)
        return padded_tokenized, target

    def step(self):
        if self.len + self.step_size <= self.max_len:
            self.len += self.step_size
        else:
            self.len = self.max_len
    
    def seq_len_range(self):
        return self.min_len, self.max_len

class IMDb(Dataset):
    def __init__(self, train, min_len=1, max_len=1000, warmup_epochs=0):
        if warmup_epochs < 1:
            self.min_len = max_len
        else:
            self.min_len = min_len
        self.max_len = max_len
        self.len = self.min_len
        self.step_size = (self.max_len - self.min_len) // (warmup_epochs + 1)
        self.data = load_dataset('imdb')['train' if train else 'test']
        
    def tokenizer(self, text):
        """
        Tokenizes the input text for IMDb dataset.
        0: CLS
        1: PAD
        """
        return [0] + [2 + ord(c) for c in text]
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        item = self.data[idx]
        tokenized = torch.tensor(self.tokenizer(item['text']), dtype=torch.long)
        target = torch.tensor(item['label'], dtype=torch.long)
        padded_tokenized = torch.nn.functional.pad(tokenized, (0, self.len - tokenized.size(0)), value=1)  # Pad with PAD token (1)
        return padded_tokenized, target
    
    def step(self):
        if self.len + self.step_size <= self.max_len:
            self.len += self.step_size
        else:
            self.len = self.max_len
    
    def seq_len_range(self):
        return self.min_len, self.max_len

class TinyShakespeare(Dataset):
    def __init__(self, train, tokenizer, min_len=1, max_len=1000, warmup_epochs=0):
        if warmup_epochs < 1:
            self.min_len = max_len
        else:
            self.min_len = min_len
        self.max_len = max_len
        self.len = self.min_len
        self.step_size = (self.max_len - self.min_len) // (warmup_epochs + 1)
        self.data = load_dataset('tiny_shakespeare')['train' if train else 'test']['text'][0].split(' ')
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
    
    def __len__(self):
        return len(self.data) // 8
    
    def __getitem__(self, idx):
        start_idx = torch.randint(0, len(self.data) - self.len - 1, (1,)).item()
        end_idx = start_idx + self.len
        item = ' '.join(self.data[start_idx:end_idx])
        tokenized = torch.tensor(self.tokenizer(item)['input_ids'], dtype=torch.long)
        pad_token_id = self.tokenizer.pad_token_id if self.tokenizer.pad_token_id is not None else self.tokenizer.eos_token_id
        padded_tokenized = torch.nn.functional.pad(tokenized, (0, self.len - tokenized.size(0)), value=pad_token_id)
        return padded_tokenized[:-1], padded_tokenized[1:]
    
    def step(self):
        if self.len + self.step_size <= self.max_len:
            self.len += self.step_size
        else:
            self.len = self.max_len
    
    def seq_len_range(self):
        return self.min_len, self.max_len

class LAMBADA(Dataset):
    def __init__(self, split, tokenizer, min_len=1, max_len=1000, warmup_epochs=0):
        """
        splits: ["train", "validation", "test"]
        """
        if warmup_epochs < 1:
            self.min_len = max_len
        else:
            self.min_len = min_len
        self.max_len = max_len
        self.len = self.min_len
        self.step_size = (self.max_len - self.min_len) // (warmup_epochs + 1)
        self.data = load_dataset('lambada', split=split)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        # ignore token id = -100
        item = self.data[idx]['text'].strip().split()
        context = self.tokenizer(' '.join(item[:-1]), add_special_tokens=False)['input_ids']
        label = self.tokenizer(" " + item[-1], add_special_tokens=False)['input_ids']
        full_context = torch.tensor(context + label[:-1], dtype=torch.long)
        full_label = torch.tensor([-100] * len(context[1:]) + label, dtype=torch.long)
        full_context = full_context[-self.len:]
        full_label = full_label[-self.len:]
        pad_token_id = self.tokenizer.pad_token_id if self.tokenizer.pad_token_id is not None else self.tokenizer.eos_token_id
        padded_context = torch.nn.functional.pad(full_context, (0, self.len - full_context.size(0)), value=pad_token_id)
        padded_label = torch.nn.functional.pad(full_label, (0, self.len - full_label.size(0)), value=-100)
        return padded_context, padded_label
    
    def step(self):
        if self.len + self.step_size <= self.max_len:
            self.len += self.step_size
        else:
            self.len = self.max_len
    
    def seq_len_range(self):
        return self.min_len, self.max_len

class ThePile(Dataset):
    def __init__(self, split, tokenizer, min_len=1, max_len=1000, warmup_epochs=0):
        """
        splits: ["train", "validation", "test"]
        """
        if warmup_epochs < 1:
            self.min_len = max_len
        else:
            self.min_len = min_len
        self.max_len = max_len
        self.len = self.min_len
        self.step_size = (self.max_len - self.min_len) // (warmup_epochs + 1)
        self.data = load_dataset('monology/pile-uncopyrighted', split=split, streaming=split=="train")
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, idx):
        item = self.data[idx]['text']
        tokenized = torch.tensor(self.tokenizer(item, add_special_tokens=False)['input_ids'], dtype=torch.long)
        pad_token_id = self.tokenizer.pad_token_id if self.tokenizer.pad_token_id is not None else self.tokenizer.eos_token_id
        padded_tokenized = torch.nn.functional.pad(tokenized, (0, self.len - tokenized.size(0)), value=pad_token_id)
        return padded_tokenized[:-1], padded_tokenized[1:]
    
    def step(self):
        if self.len + self.step_size <= self.max_len:
            self.len += self.step_size
        else:
            self.len = self.max_len
    
    def seq_len_range(self):
        return self.min_len, self.max_len

def initialize_dataset(name, *args, **kwargs):
    dataset_class = getattr(sys.modules[__name__], name, None)
    return dataset_class(*args, **kwargs)



import sys
from torch.optim.lr_scheduler import *
from transformers import get_cosine_schedule_with_warmup

class CosineAnnealingLRWithWarmup():
    def __init__(self, optimizer, num_warmup_steps, num_training_steps, num_cycles=0.5, last_epoch=-1):
        self.scheduler = get_cosine_schedule_with_warmup(
            optimizer,
            num_warmup_steps=num_warmup_steps,
            num_training_steps=num_training_steps,
            num_cycles=num_cycles,
            last_epoch=last_epoch
        )

    def step(self):
        self.scheduler.step()

    def get_last_lr(self):
        return self.scheduler.get_last_lr()

def initialize_scheduler(name, *args, **kwargs):
    scheduler_class = getattr(sys.modules[__name__], name, None)
    return scheduler_class(*args, **kwargs)